//		Copyright © 1999–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Screen saver class.

#include <climits>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <GraphicsDefs.h>
#include <image.h>
#include <List.h>
#include <Message.h>
#include <OS.h>
#include <Point.h>
#include <Polygon.h>
#include <Rect.h>
#include <Screen.h>
#include <ScreenSaver.h>
#include <TextView.h>
#include <View.h>

#include "MapSaver.h"
#include "RandomColor.h"
#include "RandomNumbers.h"

const rgb_color kBlack = { 0, 0, 0, 0 };

// Screen saver interface.
BScreenSaver *
instantiate_screen_saver(BMessage *pMsg, image_id id)
{
	return new MapSaver(pMsg, id);
}

MapSaver::MapSaver(BMessage *pMsg, image_id id)
:	inherited(pMsg, id),
	mDelay(0),
	mContours()
{
	srand((long)system_time());
	
	ReadContours("zcat /boot/optional/goodies/maps/coastline.vec.Z");
	ReadContours("zcat /boot/optional/goodies/maps/island.vec.Z");
	ReadContours("zcat /boot/optional/goodies/maps/lakes.vec.Z");
	ReadContours("zcat /boot/optional/goodies/maps/river.vec.Z");
//	ReadContours("zcat /boot/optional/goodies/maps/country.vec.Z");
//	ReadContours("zcat /boot/optional/goodies/maps/states.vec.Z");
}

MapSaver::~MapSaver(void)
{
	while (!mContours.IsEmpty()) {
		delete [] (short *)mContours.RemoveItem(mContours.CountItems() - 1);
	}
}

void
MapSaver::Draw(BView *pView, int32 frame)
{
	const int32 nContours = mContours.CountItems();
	const BRect bounds = pView->Bounds();
	const BRect mapExtent(-10800.0, -5400.0, 10800.0, 5400.0);

	static BRect mapArea;
	if (frame % (nContours + mDelay) == 0) {
		// Select a new area to display.
		pView->SetLowColor(kBlack);
		pView->FillRect(bounds, B_SOLID_LOW);

		pView->SetHighColor(RandomColor(0.5, 1.0, 0.5, 1.0));

		const float mapWidth = mapExtent.Width();
		const float mapHeight = mapExtent.Height();
		const float factor = randomFloat(0.2, 0.8);
		const float areaWidth = mapWidth * factor;
		const float areaHeight = mapHeight * factor;
		const BPoint areaLoc =
			mapExtent.LeftTop() + BPoint(randomFloat(0.0, mapWidth-areaWidth),
												  randomFloat(0.0, mapHeight-areaHeight));
		mapArea = BRect(areaLoc, areaLoc + BPoint(areaWidth, areaHeight));
	}

	const int32 index = frame % (nContours + mDelay);
	if (const short *pContour = (const short *)mContours.ItemAt(index)) {
		BPolygon polygon;
		const int32 nPoints = (int32)pContour[0];
		for (int32 i = 1; i < nPoints; ++i) {
			const BPoint p((float)pContour[1 + 2*i], -(float)pContour[2 + 2*i]);
			polygon.AddPoints(&p, 1);
		}
		polygon.MapTo(mapArea, bounds);
		
		pView->StrokePolygon(&polygon, false);
	}
		
	pView->Sync();
}

status_t
MapSaver::InitCheck(void)
{
	// We need at least one contour.
	if (mContours.IsEmpty()) {
		return B_ERROR;
	}
	return B_OK;
}

void
MapSaver::StartConfig(BView *pConfigView)
{
	/*const*/ BRect frame = pConfigView->Bounds().InsetBySelf(10.0, 10.0);
	BTextView *pCaption =
		new BTextView(frame, "MapSaver", frame.OffsetToCopy(B_ORIGIN),
							B_FOLLOW_ALL, B_WILL_DRAW);
	pCaption->SetText("MapSaver\n"
							"\n"
							"\t© 1999  Jens Kilian <jjk@acm.org>\n"
							"\n"
							"MapSaver comes with ABSOLUTELY NO WARRANTY; "
							"it is free software, and you are welcome "
							"to redistribute it under certain conditions.  "
							"See the GNU General Public License for details.");
	pConfigView->AddChild(pCaption);
	pCaption->SetViewColor(pConfigView->ViewColor());
	pCaption->SetStylable(true);
	pCaption->SetFontAndColor(be_plain_font);
	const char *pText = pCaption->Text();
	pCaption->SetFontAndColor(0, strchr(pText, '\n') - pText, be_bold_font);
	pCaption->MakeEditable(FALSE);
	pCaption->MakeSelectable(FALSE);
}

status_t
MapSaver::StartSaver(BView * /*pView*/, bool preview)
{
	SetTickSize(preview ? 100000 : 10000);	// 0.1/0.01 seconds
	mDelay = preview ? 10 : 500;				// delay 1/5 seconds
	return InitCheck();
}

void
MapSaver::ReadContours(const char *pName)
{
	FILE *pFile = popen(pName, "r");
	if (pFile) {
		while (short *pContour = ReadContour(pFile)) {
			mContours.AddItem(pContour);
		}
		pclose(pFile);
	}
}

short *
MapSaver::ReadContour(FILE *pFile)
{
	int nPoints;
	if (fscanf(pFile, "%*d %d\n", &nPoints) == EOF) {
		return 0;
	}

	short *pContour = new short[1 + 2*nPoints];
	pContour[0] = (short)nPoints;
	for (int i = 0; i < nPoints; ++i) {
		int x, y;
		fscanf(pFile, "%d %d\n", &x, &y);
		pContour[1 + 2*i] = (short)x;
		pContour[2 + 2*i] = (short)y;
	}

	return pContour;
}
