//		Copyright © 1999–2000  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Screen saver class.

#include <cstdlib>
#include <cstring>

#include <GraphicsDefs.h>
#include <image.h>
#include <Message.h>
#include <OS.h>
#include <Point.h>
#include <Rect.h>
#include <ScreenSaver.h>
#include <TextView.h>
#include <View.h>

#include "PentoSaver.h"
#include "RandomColor.h"
#include "RandomNumbers.h"

const rgb_color kBlack = { 0, 0, 0, 0 };

const float kMinimalTileSize = 12;
const unsigned int kDefaultTiles = 24;

const int kMinRemoved = 1;
const int kMaxRemoved = 5;

// Pentomino tiles - FILNPTUVWXYZ.
const unsigned int kNumTiles = 12;
const unsigned int kTileArea = 5;
const unsigned int kNumOrientations = 8;

int kTiles[kNumTiles][kTileArea][2] = {
	{	{ 0,-1}, { 1,-1}, {-1, 0}, { 0, 0}, { 0, 1}	},
	{	{-2, 0}, {-1, 0}, { 0, 0}, { 1, 0}, { 2, 0}	},
	{	{ 0,-1}, { 0, 0}, { 0, 1}, { 0, 2}, { 1, 2}	},
	{	{ 1,-1}, { 0, 0}, { 1, 0}, { 0, 1}, { 0, 2}	},
	{	{ 0,-1}, { 1,-1}, { 0, 0}, { 1, 0}, { 0, 1}	},
	{	{-1,-1}, { 0,-1}, { 1,-1}, { 0, 0}, { 0, 1}	},
	{	{-1,-1}, { 1,-1}, {-1, 0}, { 0, 0}, { 1, 0}	},
	{	{ 0,-2}, { 0,-1}, { 0, 0}, { 1, 0}, { 2, 0}	},
	{	{ 0,-2}, { 0,-1}, { 1,-1}, { 1, 0}, { 2, 0}	},
	{	{ 0,-1}, {-1, 0}, { 0, 0}, { 1, 0}, { 0, 1}	},
	{	{ 0,-1}, { 0, 0}, { 1, 0}, { 0, 1}, { 0, 2}	},
	{	{-1,-1}, { 0,-1}, { 0, 0}, { 0, 1}, { 1, 1}	}
};

// Screen saver interface.
BScreenSaver *
instantiate_screen_saver(BMessage *pMsg, image_id id)
{
	return new PentoSaver(pMsg, id);
}

PentoSaver::PentoSaver(BMessage *pMsg, image_id id)
:	inherited(pMsg, id),
	mTileSize(0),
	mWidth(0),
	mHeight(0),
	mpArray(0),
	mpLocations(0),
	mPlaced(),
	mFull(false)
{
	srand((long)system_time());
}

PentoSaver::~PentoSaver(void)
{
	delete [] mpArray;
	mpArray = 0;
	delete [] mpLocations;
	mpLocations = 0;
	while (!mPlaced.IsEmpty()) {
		delete (Location *)(mPlaced.RemoveItem(mPlaced.CountItems() - 1));
	}
}

void
PentoSaver::Draw(BView *pView, int32 frame)
{
	if (!pView || !mpArray || !mpLocations) {
		return;
	}
	
	if (frame == 0) {
		// Erase entire screen.
		WipeScreen(pView);
	}

	if (mFull) {
		// No free location found in previous frame - make some space.
		
		// Find all tiles which border some empty space.
		BList freeTiles;
		for (int i = mPlaced.CountItems(); i > 0;) {
			Location *pLoc = (Location *)(mPlaced.ItemAt(--i));
			if (pLoc && TileFree(pLoc->tile,pLoc->orientation,pLoc->x,pLoc->y)) {
				freeTiles.AddItem(pLoc);
			}
		}

		const int nFound = freeTiles.CountItems();
		if (nFound < 1) {
			// A perfect tiling.  Too bad we'll have to wipe it to continue!
			WipeScreen(pView);
		
		} else {
			// Remove some of the less than perfectly placed tiles.
			const int toRemove =
				kMinRemoved + randomInteger(kMaxRemoved-kMinRemoved);
			for (int i = 0; i < toRemove && !freeTiles.IsEmpty(); ++i) {
				const int index = randomInteger(freeTiles.CountItems());
				Location *pRemoved = (Location *)(freeTiles.RemoveItem(index));
				mPlaced.RemoveItem(pRemoved);
				DrawTile(pView,
							pRemoved->tile,
							pRemoved->orientation,
							pRemoved->x,
							pRemoved->y,
							true);
				delete pRemoved;
			}
		}
	}
	
	// Find all possible tile placements.
	int nLocations = 0;
	for (unsigned int tile = 0; tile < kNumTiles; ++tile) {
		for (unsigned int orientation = 0;
			  orientation < kNumOrientations;
			  ++orientation) {
			for (int x = 0; x < mWidth; ++x) {
				for (int y = 0; y < mHeight; ++y) {
					if (TileFits(tile, orientation, x, y)) {
						// Remember this location.
						mpLocations[nLocations].tile = tile;
						mpLocations[nLocations].orientation = orientation;
						mpLocations[nLocations].x = x;
						mpLocations[nLocations].y = y;
						++nLocations;
					}
				}
			}
		}
	}

	if (nLocations > 0) {
		mFull = false;
		// Pick a random location and draw.
		const int loc = randomInteger(nLocations);

		pView->SetHighColor(RandomColor(0.5, 1.0, 0.5, 1.0));
		DrawTile(pView,
					mpLocations[loc].tile,
					mpLocations[loc].orientation,
					mpLocations[loc].x,
					mpLocations[loc].y);
		
		mPlaced.AddItem(new Location(mpLocations[loc]));
		
	} else {
		// No more room.
		mFull = true;
	}
	
	pView->Sync();
}

void
PentoSaver::StartConfig(BView *pConfigView)
{
	/*const*/ BRect frame = pConfigView->Bounds().InsetBySelf(10.0, 10.0);
	BTextView *pCaption =
		new BTextView(frame, "PentoSaver", frame.OffsetToCopy(B_ORIGIN),
							B_FOLLOW_ALL, B_WILL_DRAW);
	pCaption->SetText("PentoSaver\n"
							"\n"
							"\t© 1999  Jens Kilian <jjk@acm.org>\n"
							"\n"
							"PentoSaver comes with ABSOLUTELY NO WARRANTY; "
							"it is free software, and you are welcome "
							"to redistribute it under certain conditions.  "
							"See the GNU General Public License for details.");
	pConfigView->AddChild(pCaption);
	pCaption->SetViewColor(pConfigView->ViewColor());
	pCaption->SetStylable(true);
	pCaption->SetFontAndColor(be_plain_font);
	const char *pText = pCaption->Text();
	pCaption->SetFontAndColor(0, strchr(pText, '\n') - pText, be_bold_font);
	pCaption->MakeEditable(FALSE);
	pCaption->MakeSelectable(FALSE);
}

status_t
PentoSaver::StartSaver(BView *pView, bool preview)
{
	SetTickSize(preview ? 1000000 : 100000);	// 1/0.1 second

	// Try default number of tiles, reduce until minimal size is guaranteed.
	const BRect bounds = pView->Bounds();
	const float width = bounds.Width();
	const float height = bounds.Height();

	int nTiles = kDefaultTiles;
	while (nTiles > 0) {
		const float hSize = width / nTiles;
		const float vSize = height / nTiles;
		if (hSize > kMinimalTileSize && vSize > kMinimalTileSize) {
			mTileSize = (hSize > vSize) ? vSize : hSize;
			break;
		}
		--nTiles;
	}

	if (nTiles == 0) {
		// View too small.
		return B_ERROR;
	}

	// Initialize bit array.
	mWidth = (int)(width / mTileSize);
	mHeight = (int)(height / mTileSize);
	const int nBits = mWidth * mHeight;
	delete [] mpArray;
	mpArray = new bool[nBits];
	delete [] mpLocations;
	mpLocations = new Location[nBits * kNumTiles * kNumOrientations];
	
	return B_OK;
}

// Erase entire screen.
void
PentoSaver::WipeScreen(BView *pView)
{
	const BRect bounds = pView->Bounds();
	pView->SetLowColor(kBlack);
	pView->FillRect(bounds, B_SOLID_LOW);
	
	// Erase array.
	for (int i = mWidth * mHeight; i > 0;) {
		mpArray[--i] = false;
	}
	mFull = false;
}

// Check if there is space in the array for a given pentomino.
bool
PentoSaver::TileFits(unsigned int tile, unsigned int orientation, int x, int y)
{
	const int dx = (orientation & 0x1) ? -1 : 1;
	const int dy = (orientation & 0x2) ? -1 : 1;
	const int mirror = (orientation & 0x4) != 0;
	
	for (unsigned int i = 0; i < kTileArea; ++i) {
		const int xx = x + dx * kTiles[tile][i][mirror];
		const int yy = y + dy * kTiles[tile][i][!mirror];
		if (xx < 0 || xx >= mWidth
			 || yy < 0 || yy >= mHeight
			 || mpArray[xx + mWidth * yy]) {
			return false;
		}
	}
	
	return true;
}

// Check if there is space in the array around a pentomino which was already
// placed on the screen.
bool
PentoSaver::TileFree(unsigned int tile, unsigned int orientation, int x, int y)
{
	const int dx = (orientation & 0x1) ? -1 : 1;
	const int dy = (orientation & 0x2) ? -1 : 1;
	const int mirror = (orientation & 0x4) != 0;
	
	for (unsigned int i = 0; i < kTileArea; ++i) {
		const int xx = x + dx * kTiles[tile][i][mirror];
		const int yy = y + dy * kTiles[tile][i][!mirror];

		if (xx-1 >= 0 && xx-1 < mWidth
			 && yy >= 0 && yy < mHeight
			 && !mpArray[xx-1 + mWidth * yy]) {
			return true;
		}

		if (xx+1 >= 0 && xx+1 < mWidth
			 && yy >= 0 && yy < mHeight
			 && !mpArray[xx+1 + mWidth * yy]) {
			return true;
		}

		if (xx >= 0 && xx < mWidth
			 && yy-1 >= 0 && yy-1 < mHeight
			 && !mpArray[xx + mWidth * (yy-1)]) {
			return true;
		}

		if (xx >= 0 && xx < mWidth
			 && yy+1 >= 0 && yy+1 < mHeight
			 && !mpArray[xx + mWidth * (yy+1)]) {
			return true;
		}
	}
	
	return false;
}

// Draw a pentomino.
void
PentoSaver::DrawTile(BView *pView,
							unsigned int tile, unsigned int orientation, int x, int y,
							bool remove)
{
	const int dx = (orientation & 0x1) ? -1 : 1;
	const int dy = (orientation & 0x2) ? -1 : 1;
	const int mirror = (orientation & 0x4) != 0;

	int hLines[kTileArea+1][kTileArea];
	int vLines[kTileArea][kTileArea+1];
	memset(hLines, 0, sizeof(hLines));
	memset(vLines, 0, sizeof(vLines));
	
	for (unsigned int i = 0; i < kTileArea; ++i) {
		const int tx = dx * kTiles[tile][i][mirror];
		const int ty = dy * kTiles[tile][i][!mirror];

		const int xx = x + tx;
		const int yy = y + ty;
		mpArray[xx + mWidth * yy] = !remove;
		BRect rect(xx * mTileSize, yy * mTileSize,
					  (xx+1) * mTileSize - 1, (yy+1) * mTileSize - 1);
		pView->FillRect(rect, remove ? B_SOLID_LOW : B_SOLID_HIGH);
		
		hLines[ty + 2][tx + 2] ^= 1;
		hLines[ty + 3][tx + 2] ^= 1;
		vLines[ty + 2][tx + 2] ^= 1;
		vLines[ty + 2][tx + 3] ^= 1;
	}

	pView->BeginLineArray(20);
	for (unsigned int i = 0; i <= kTileArea; ++i) {
		for (unsigned int j = 0; j < kTileArea; ++j) {
			if (hLines[i][j]) {
				pView->AddLine(BPoint((x-2+j) * mTileSize, (y-2+i) * mTileSize),
									BPoint((x-1+j) * mTileSize, (y-2+i) * mTileSize),
									kBlack);
			}
			if (vLines[j][i]) {
				pView->AddLine(BPoint((x-2+i) * mTileSize, (y-2+j) * mTileSize),
									BPoint((x-2+i) * mTileSize, (y-1+j) * mTileSize),
									kBlack);
			}
		}
	}
	pView->EndLineArray();
}
